package header

import (
	"bytes"
	"fmt"
	"io"
	"nproject.io/gitlab/root/dns/binary"
	"nproject.io/gitlab/root/dns/packet/code"
)

const (
	HeaderLengthBytes = 12
)

type Header struct {
	// ID
	ID int16

	// OPCode
	OPCode code.Op

	// QuestionCount is QDCount
	QuestionCount uint16

	// ResponseCount is ANCount
	ResponseCount uint16

	// NameServerCount is NSCount
	NameServerCount uint16

	// AdditionalAnswerCount is ARCount
	AdditionalAnswerCount uint16

	// ResponseCode
	ResponseCode code.R

	// QR
	IsQuestion bool
	// AA - response only
	IsAuthoritative bool
	// TC
	IsTruncated bool
	// RD
	IsRecursionRequired bool
	// RA - response only
	IsRecursionAvailable bool
}

func ParseHeader(reader io.ReadSeeker) (*Header, error) {
	header := &Header{}

	if id, err := binary.ReadInt16(reader); err == nil {
		header.ID = id
	} else {
		return nil, err
	}

	if flags, err := binary.ReadInt16(reader); err == nil {
		header.IsQuestion = (uint16(flags) & 0x8000) == 0

		header.OPCode = code.Op((uint16(flags) >> 11) & 0xF)
		header.IsAuthoritative = (uint16(flags) >> 10 & 0x01) == 1
		header.IsTruncated = (uint16(flags) >> 9 & 0x01) == 1
		header.IsRecursionRequired = (uint16(flags) >> 8 & 0x01) == 1
		header.IsRecursionAvailable = (uint16(flags) >> 7 & 0x01) == 1

		header.ResponseCode = code.R(uint16(flags) & 0xF)
	} else {
		return nil, err
	}

	if questionCount, err := binary.ReadInt16(reader); err == nil {
		header.QuestionCount = uint16(questionCount)
	} else {
		return nil, err
	}

	if answerCount, err := binary.ReadInt16(reader); err == nil {
		header.ResponseCount = uint16(answerCount)
	} else {
		return nil, err
	}

	if nameServerCount, err := binary.ReadInt16(reader); err == nil {
		header.NameServerCount = uint16(nameServerCount)
	} else {
		return nil, err
	}

	if additionalCount, err := binary.ReadInt16(reader); err == nil {
		header.AdditionalAnswerCount = uint16(additionalCount)
	} else {
		return nil, err
	}

	return header, nil
}

func (header *Header) String() string {
	builder := bytes.NewBuffer(nil)
	builder.WriteString("Packet is ")
	if header.IsQuestion {
		builder.WriteString("a question ")
		builder.WriteString(fmt.Sprintf("TC(%v), RD(%v) with %d entries:\n", header.IsTruncated, header.IsRecursionRequired, header.QuestionCount))
	} else {
		builder.WriteString("a response ")
		builder.WriteString(fmt.Sprintf("AA(%v) TC(%v), RD(%v), RA(%v) :\n", header.IsAuthoritative, header.IsTruncated, header.IsRecursionRequired, header.IsRecursionAvailable))
	}
	return builder.String()
}

func (header *Header) flag() uint16 {
	flag := uint16(0)
	flag |= uint16(binary.BoolToByte(!header.IsQuestion)) << 15
	flag |= uint16(header.OPCode) << 11
	flag |= uint16(binary.BoolToByte(header.IsAuthoritative)) << 10
	flag |= uint16(binary.BoolToByte(header.IsTruncated)) << 9
	flag |= uint16(binary.BoolToByte(header.IsRecursionRequired)) << 8
	flag |= uint16(binary.BoolToByte(header.IsRecursionAvailable)) << 7
	flag |= uint16(header.ResponseCode)
	return flag
}

func (header *Header) Write(writer binary.Writer) {
	_ = binary.WriteInt16(writer, header.ID)
	_ = binary.WriteInt16(writer, int16(header.flag()))
	_ = binary.WriteInt16(writer, int16(header.QuestionCount))
	_ = binary.WriteInt16(writer, int16(header.ResponseCount))
	_ = binary.WriteInt16(writer, int16(header.NameServerCount))
	_ = binary.WriteInt16(writer, int16(header.AdditionalAnswerCount))
}
