package packet

import (
	"bytes"
	"io"
	Header2 "nproject.io/gitlab/root/dns/packet/header"
	"nproject.io/gitlab/root/dns/packet/pointer"
	"nproject.io/gitlab/root/dns/packet/question"
	"nproject.io/gitlab/root/dns/packet/record"
	"strings"
)

// Packet is a dns packet
// https://datatracker.ietf.org/doc/html/rfc1035#section-4.1.1
type Packet struct {
	Header Header2.Header

	Question   []question.Question
	Response   []record.Record
	Authority  []record.Record
	Additional []record.Record
}

func Parse(reader io.ReadSeeker) (*Packet, error) {
	packet := &Packet{}

	if header, err := Header2.ParseHeader(reader); err == nil {
		packet.Header = *header
	} else {
		return nil, err
	}

	if packet.Header.QuestionCount > 0 {
		packet.Question = make([]question.Question, packet.Header.QuestionCount)
		for currentQuestion := uint16(0); currentQuestion < packet.Header.QuestionCount; currentQuestion++ {
			if readQuestion, err := question.ParseQuestion(reader); err == nil {
				packet.Question[currentQuestion] = *readQuestion
			} else {
				return nil, err
			}
		}
	}

	if packet.Header.ResponseCount > 0 {
		packet.Response = make([]record.Record, packet.Header.ResponseCount)
		for currentResponse := uint16(0); currentResponse < packet.Header.ResponseCount; currentResponse++ {
			if response, err := record.ParseRecord(reader); err == nil {
				packet.Response[currentResponse] = *response
			} else {
				return nil, err
			}
		}
	}

	if packet.Header.NameServerCount > 0 {
		packet.Authority = make([]record.Record, packet.Header.NameServerCount)
		for currentAuthority := uint16(0); currentAuthority < packet.Header.NameServerCount; currentAuthority++ {
			if authority, err := record.ParseRecord(reader); err == nil {
				packet.Authority[currentAuthority] = *authority
			} else {
				return nil, err
			}
		}
	}

	if packet.Header.AdditionalAnswerCount > 0 {
		packet.Additional = make([]record.Record, packet.Header.AdditionalAnswerCount)
		for currentAdditional := uint16(0); currentAdditional < packet.Header.AdditionalAnswerCount; currentAdditional++ {
			if additional, err := record.ParseRecord(reader); err == nil {
				packet.Additional[currentAdditional] = *additional
			} else {
				return nil, err
			}
		}
	}

	return packet, nil
}

func (packet *Packet) Bytes() []byte {
	pointerRegister := pointer.New()
	writer := bytes.NewBuffer(nil)

	packet.Header.Write(writer)
	for _, currentQuestion := range packet.Question {
		currentQuestion.Write(writer,
			pointerRegister)
	}
	for _, currentResponse := range packet.Response {
		currentResponse.Write(writer,
			pointerRegister)
	}
	for _, currentAuthority := range packet.Authority {
		currentAuthority.Write(writer,
			pointerRegister)
	}
	for _, currentAdditional := range packet.Additional {
		currentAdditional.Write(writer,
			pointerRegister)
	}

	return writer.Bytes()
}

func (packet *Packet) String() string {
	builder := &strings.Builder{}

	builder.WriteString(packet.Header.String())

	if packet.Header.IsQuestion {
		for _, currentQuestion := range packet.Question {
			builder.WriteString("- " + currentQuestion.String() + " ?")
		}
	} else {
		for _, currentAnswer := range packet.Response {
			builder.WriteString("- " + currentAnswer.String() + " !")
		}
	}

	return builder.String()
}
