package packet

import (
	"bytes"
	"github.com/stretchr/testify/require"
	"nproject.io/gitlab/root/dns/packet/header"
	"nproject.io/gitlab/root/dns/packet/name"
	"nproject.io/gitlab/root/dns/packet/query"
	"nproject.io/gitlab/root/dns/packet/question"
	"nproject.io/gitlab/root/dns/packet/record"
	"testing"
)

func TestParseAndBytes(t *testing.T) {
	type testType struct {
		name string

		givenPacket []byte

		expectedResult *Packet
		expectedError  error
	}

	for _, currentTest := range []testType{
		{
			name:        "request google.com",
			givenPacket: []byte{199, 8, 1, 0, 0, 1, 0, 0, 0, 0, 0, 0, 6, 103, 111, 111, 103, 108, 101, 3, 99, 111, 109, 0, 0, 1, 0, 1},
			expectedResult: &Packet{
				Header: header.Header{
					ID:                    -14584,
					OPCode:                0,
					QuestionCount:         1,
					ResponseCount:         0,
					NameServerCount:       0,
					AdditionalAnswerCount: 0,
					ResponseCode:          0,
					IsQuestion:            true,
					IsAuthoritative:       false,
					IsTruncated:           false,
					IsRecursionRequired:   true,
					IsRecursionAvailable:  false,
				},
				Question: []question.Question{
					{
						Name: name.Name{
							Name: []string{
								"google",
								"com",
							},
						},
						Type:  query.QType(query.TypeA),
						Class: query.ClassIN,
					},
				},
			},
			expectedError: nil,
		},
		{
			name:        "response google.com",
			givenPacket: []byte{1, 1, 129, 128, 0, 1, 0, 1, 0, 0, 0, 0, 6, 103, 111, 111, 103, 108, 101, 3, 99, 111, 109, 0, 0, 1, 0, 1, 192, 12, 0, 1, 0, 1, 0, 0, 0, 72, 0, 4, 142, 250, 201, 174},
			expectedResult: &Packet{
				Header: header.Header{
					ID:                   257,
					QuestionCount:        1,
					ResponseCount:        1,
					IsRecursionRequired:  true,
					IsRecursionAvailable: true,
				},
				Question: []question.Question{
					{
						Name: name.Name{
							Name: []string{"google", "com"},
						},
						Type:  query.QType(query.TypeA),
						Class: query.ClassIN,
					},
				},
				Response: []record.Record{
					{
						Name: name.Name{
							Name: []string{"google", "com"},
						},
						Type:               query.TypeA,
						Class:              query.ClassIN,
						TTL:                72,
						ResourceDataLength: 4,
						ResourceData:       []uint8{0x8e, 0xfa, 0xc9, 0xae},
					},
				},
			}, // 142.250.201.174 and 2a00:1450:4007:81a::200e
			expectedError: nil,
		},
		{
			name:          "response registry.nproject.io",
			givenPacket:   []byte{1, 1, 129, 128, 0, 1, 0, 1, 0, 0, 0, 0, 8, 114, 101, 103, 105, 115, 116, 114, 121, 8, 110, 112, 114, 111, 106, 101, 99, 116, 2, 105, 111, 0, 0, 1, 0, 1, 192, 12, 0, 1, 0, 1, 0, 0, 1, 206, 0, 4, 86, 242, 74, 228},
			expectedError: nil,
			expectedResult: &Packet{
				Header: header.Header{
					ID:                   257,
					OPCode:               0,
					QuestionCount:        1,
					ResponseCount:        1,
					IsRecursionRequired:  true,
					IsRecursionAvailable: true,
				},
				Question: []question.Question{
					{
						Name: name.Name{
							Name: []string{
								"registry",
								"nproject",
								"io",
							},
						},
						Type:  query.QType(query.TypeA),
						Class: query.ClassIN,
					},
				},
				Response: []record.Record{
					{
						Name: name.Name{
							Name: []string{
								"registry",
								"nproject",
								"io",
							},
						},
						Type:               query.TypeA,
						Class:              query.ClassIN,
						TTL:                462,
						ResourceDataLength: 4,
						ResourceData:       []byte{86, 242, 74, 228},
					},
				},
			},
		},
		{
			name:        "true response detectportal.firefox.com",
			givenPacket: []byte{73, 178, 129, 128, 0, 1, 0, 3, 0, 0, 0, 1, 12, 100, 101, 116, 101, 99, 116, 112, 111, 114, 116, 97, 108, 7, 102, 105, 114, 101, 102, 111, 120, 3, 99, 111, 109, 0, 0, 1, 0, 1, 192, 12, 0, 5, 0, 1, 0, 0, 0, 19, 0, 30, 12, 100, 101, 116, 101, 99, 116, 112, 111, 114, 116, 97, 108, 4, 112, 114, 111, 100, 6, 109, 111, 122, 97, 119, 115, 3, 110, 101, 116, 0, 192, 54, 0, 5, 0, 1, 0, 0, 0, 50, 0, 41, 4, 112, 114, 111, 100, 12, 100, 101, 116, 101, 99, 116, 112, 111, 114, 116, 97, 108, 4, 112, 114, 111, 100, 8, 99, 108, 111, 117, 100, 111, 112, 115, 6, 109, 111, 122, 103, 99, 112, 192, 79, 192, 96, 0, 1, 0, 1, 0, 0, 2, 41, 0, 4, 34, 107, 221, 82, 0, 0, 41, 255, 214, 0, 0, 0, 0, 0, 0},
		},
	} {
		t.Run(currentTest.name,
			func(t *testing.T) {
				// given

				// when
				actualResult, actualError := Parse(bytes.NewReader(currentTest.givenPacket))

				// then
				require.Equal(t, currentTest.expectedError, actualError)
				require.Equal(t, currentTest.expectedResult, actualResult)

				if currentTest.expectedError == nil {
					require.Equal(t, currentTest.givenPacket, actualResult.Bytes())
				}
			},
		)
	}
}
