package record

import (
	"io"
	"nproject.io/gitlab/root/dns/binary"
	"nproject.io/gitlab/root/dns/packet"
	"nproject.io/gitlab/root/dns/packet/header"
	"nproject.io/gitlab/root/dns/packet/name"
	"nproject.io/gitlab/root/dns/packet/pointer"
	"nproject.io/gitlab/root/dns/packet/query"
	"nproject.io/gitlab/root/dns/packet/record/extended"
	"strings"
)

type Record struct {
	Name name.Name

	Type  query.Type
	Class query.Class
	TTL   int32

	ResourceDataLength int16
	ResourceData       []byte
}

func ParseRecord(reader io.ReadSeeker) (*Record, error) {
	response := &Record{}

	if nameInstance, err := name.Parse(reader); err == nil {
		response.Name = *nameInstance
	} else {
		return nil, err
	}

	if _type, err := binary.ReadInt16(reader); err == nil {
		response.Type = query.Type(_type)
	} else {
		return nil, err
	}
	if class, err := binary.ReadInt16(reader); err == nil {
		response.Class = query.Class(class)
	} else {
		return nil, err
	}

	if ttl, err := binary.ReadInt32(reader); err == nil {
		response.TTL = ttl
	} else {
		return nil, err
	}

	if responseDataLength, err := binary.ReadInt16(reader); err == nil {
		response.ResourceDataLength = responseDataLength
	} else {
		return nil, err
	}

	var resourceData = make([]byte, response.ResourceDataLength)
	if response.ResourceDataLength > 0 {
		if readLength, err := reader.Read(resourceData); err == nil {
			if readLength == int(response.ResourceDataLength) {
				response.ResourceData = resourceData
			} else {
				return nil, packet.NotEnoughDataError
			}
		} else {
			return nil, err
		}
	}

	return response, nil
}

func (record *Record) Write(writer binary.Writer,
	pointerRegister pointer.Register) {
	record.Name.Write(writer,
		pointerRegister)

	_ = binary.WriteInt16(writer, int16(record.Type))
	_ = binary.WriteInt16(writer, int16(record.Class))
	_ = binary.WriteInt32(writer, record.TTL)
	_ = binary.WriteInt16(writer, record.ResourceDataLength)
	_, _ = writer.Write(record.ResourceData)
}

func (record *Record) String() string {
	buffer := &strings.Builder{}

	buffer.WriteString(record.Name.String())

	return buffer.String()
}

func (record *Record) Get(header *header.Header) (interface{}, error) {
	switch record.Type {
	case query.TypeOPT:
		return extended.New(header,
			&record.Name,
			record.TTL,
			int16(record.Class),
			record.ResourceDataLength,
			record.ResourceData)
	default:
		return nil, packet.NotImplementedError
	}

}
